# PET regression models - Functions -------------------------
# Authors: Rahma ABID - Clestin C. KOKONENDJI - Afif MASMOUDI----------------------------------
# Date: 04/08/2020 -----------------------------------------------------


# Simulating from Poisson-Tweedie models -------------------------------
rptweedie <- function(n, mu, phi, power) {
  if(power != 3) {
    x <- rtweedie(n = n, mu = mu, phi = phi, power = power)
  }
  if(power == 3) {
    x <- rinvgauss(n = n, mean = mu, dispersion = phi)
  }
  y <- rpois(n = n, lambda = x)
  return(y)
}

# Simulating from PET models -------------------------------
rpetweedie<-function(n,mu,phi,power)
{
  S=c()
  y=rep(1,n)
  for (i in 1:n)
  {
    g=rexp(1,1)
    if(power != 0){
      y[i]=rptweedie(1,mu=g*mu[i],phi=(g^(1-power))*phi,power=power)}}
  return(y)
  
}

# Simulating from PET regression models --------------------
rpetweedie_reg <- function(beta, phi,power,X, offset = NULL) {
  eta <- X%*%beta
  mu <- as.numeric(exp(eta))
  y <-rpetweedie(n = length(mu), mu = mu, phi = phi,power=power)
  return(y)
}

# Quasi-likelihood estimation ------------------------------------------
quasi_score_beta_PET <- function(beta, lambda, y, X, offset = NULL, 
                                 sensitivity = FALSE) {
  output <- list()
  n_beta = dim(X)[2]
  phi <- lambda[1]
  power <- lambda[2]
  eta <- as.numeric(X%*%beta)
  if(!is.null(offset)) {eta <- eta + offset}
  mu <- as.numeric(exp(eta))
  mu.p <- mu^power
  den <-mu^2+ mu + phi*mu.p
  res <- (y - mu)
  res2 <- (y - mu)^2
  inv_C <- 1/den
  output$esc_beta <- colSums(mu*X*inv_C*res)
  if(sensitivity == TRUE) {
    II <- matrix(NA, ncol = n_beta, nrow = n_beta)
    for(i in 1:n_beta) {
      for(j in 1:n_beta) {
        II[i,j] <- -sum(mu*X[,i]*inv_C*mu*X[,j])
      }
    }
    output$S_beta = II
  }
  return(output)
}

# Pseudo-likelihood estimation ------------------------------------------

pearson_lambda_PET <- function(lambda, beta, y, X, offset = NULL, 
                               sensivity = FALSE) {
  output <- list()
  phi <- lambda[1]
  power <- lambda[2]
  eta <- as.numeric(X%*%beta)
  if(!is.null(offset)) {eta <- eta + offset}
  mu <- as.numeric(exp(eta))
  mu.p <- mu^power
  den <- mu^(2)+mu + phi*mu.p
  res <- (y - mu)
  res2 <- (y - mu)^2
  inv_C <- 1/den
  W_phi <- mu.p/(den^2)
  W_power <- mu.p*eta*phi/(den^2)
  esc_phi <- sum(W_phi*(res2 - den))
  esc_power <- sum(W_power*(res2 - den))
  output$esc_lambda <- c(esc_phi, esc_power)
  if(sensivity == TRUE) {
    S_phi <- - sum( (mu.p/den)^2 )
    S_phi_power <- - sum( ((mu^(2*power))*eta*phi)/den^2)
    S_power <- -sum(((mu.p*eta*phi)^2)/(den^2))
    SS <- matrix(c(S_phi, S_phi_power, S_phi_power, S_power), 2, 2)
    output$S_lambda <- SS
  }
  return(output)
}

chaser_quasi_PET <- function(reg_ini, disp_ini, Y, X, offset = NULL,
                             tol=0.0001, max_iter = 20, alpha = 1, 
                             verbose = FALSE) {
  sol_beta <- matrix(NA, max_iter,length(reg_ini))
  sol_disp <- matrix(NA, max_iter, length(disp_ini))
  sol_beta[1,] <- reg_ini
  sol_disp[1,] <- disp_ini
  for(i in 2:max_iter){
    reg_temp <- quasi_score_beta_PET(beta = reg_ini, 
                                     lambda = disp_ini, 
                                     y = Y, X = X, 
                                     sensitivity = TRUE)
    sol_beta[i,] <- as.numeric(reg_ini - solve(reg_temp$S_beta, reg_temp$esc_beta))
    reg_ini <- sol_beta[i,]
    disp_temp <- pearson_lambda_PET(lambda = disp_ini, beta = reg_ini, 
                                    y = Y, X = X, 
                                    sensivity = TRUE)
    sol_disp[i,] <- as.numeric(disp_ini - alpha*solve(disp_temp$S_lambda, disp_temp$esc_lambda))
    disp_ini <- sol_disp[i,]
    #if(verbose == TRUE) print(round(disp_ini,4))
    #tolera <- abs(c(sol_beta[i,],sol_disp[i,])  - c(sol_beta[i-1,],sol_disp[i-1,]))
    #if(all(tolera < tol) == TRUE)break
  }
  saida <- list("IterationRegression" = sol_beta, 
                "IterarionCovariance" = sol_disp, 
                "Regression" = reg_ini, 
                "Covariance" = disp_ini,
                "S_beta" = reg_temp,
                "S_lambda" = disp_temp)
  return(saida)
}  

# Glm formula ------------------------------------------

glm_GPTwee_PET<- function(formula, initial, data, tol = 1e-04,
                          max_iter = 50, alpha = 1, verbose = FALSE) {
  
  #initial <- c(ini_beta, initial)
  mf <- model.frame(formula, data = data)
  Y <- model.response(mf)
  X <- model.matrix(formula, data = data)
  n_beta <- dim(X)[2]
  reg_ini <- initial[1:n_beta]
  disp_ini <- initial[c(n_beta+1, n_beta+2)]
  output <- chaser_quasi_PET(reg_ini = reg_ini, disp_ini = disp_ini,
                             Y = Y, X = X, tol = tol, verbose = verbose,
                             max_iter = max_iter, alpha = alpha)
  
  Estimates <- c(output$Regression, output$Covariance)
  eta <- as.numeric(X%*%output$Regression)
  mu <- exp(eta)
  phi <- output$Covariance[1]
  power <- output$Covariance[2]
  mu.p <- mu^power
  den <- mu + phi*mu.p+mu^2
  S_beta_phi <- colSums(((mu.p*phi*power*X + mu*X+2*(mu^2)*X)*(mu.p))/(den^2))
  S_beta_power <- colSums(((mu.p*phi*power*X + mu*X+2*(mu^2)*X)*(mu.p*eta*phi))/(den^2))
  S_cross <- rbind(S_beta_phi,S_beta_power)
  dd <- dim(S_cross)
  S_0 <- matrix(0, ncol = dd[1], nrow = dd[2] )
  n_par <- length(initial)
  S1 <- cbind(output$S_beta$S_beta, S_0)
  S2 <- cbind(S_cross, output$S_lambda$S_lambda)
  Sensitivity = rbind(S1, S2)
  temp <- matrix(0, ncol = 2, nrow = 2)
  for(i in 1:length(Y)) {
    temp2 <- pearson_lambda_PET(lambda = output$Covariance,
                                beta = output$Regression,
                                y = Y[i], X = matrix(X[i,], 1, n_beta))
    escore <- c(temp2$esc_lambda)
    temp <- temp + escore%*%t(escore)
  }
  inv_S <- solve(-Sensitivity)
  VCOV <- inv_S%*%temp%*%inv_S
  output <- list("Estimates" = Estimates, "VCOV" = VCOV)
  return(output)
}


integrand <- function(x, y, mu, phi, power) {
  int = dpois(y, lambda = x)*dtweedie(x, mu = mu, 
                                      phi = phi, power = power)
  return(int)
}

monte_carlo <- function(integrand, n_pts, y, mu, phi, power) {
  pts <- rtweedie(n_pts, mu = mu, phi = phi, power = power)
  norma <- dtweedie(pts, mu = mu, phi = phi, power = power)
  integral <- mean(integrand(pts, y = y, mu = mu, phi = phi, 
                             power = power)/norma)
  return(integral)
}

gauss_laguerre <- function(integrand, n_pts, y, mu, phi, power) {
  pts <- gauss.quad(n_pts, kind="laguerre")
  integral <- sum(pts$weights*integrand(pts$nodes, y = y, mu = mu, 
                                        phi = phi, power = power)/
                    exp(-pts$nodes))
  return(integral)
}

gauss_laguerre_vec <- Vectorize(FUN = gauss_laguerre, vectorize.args = "y")

# Poisson-Tweedie density -----------------------

dptweedie_aux <- function(y, mu, phi, power, n_pts, method) {
  if(method == "laguerre" | y > 0) {
    pmf <- gauss_laguerre(integrand = integrand, n_pts = n_pts, 
                          y = y, mu = mu, phi = phi, power = power)
  }
  if(method == "laguerre" & y == 0) {
    v.y <- round(10*sqrt(mu + phi*mu^power),0)
    if(v.y > 1000) {v.y <- 1000}
    #print(v.y)
    y.grid <- 0:v.y
    temp <- gauss_laguerre_vec(integrand = integrand, n_pts = n_pts,
                               y = y.grid, mu = mu, phi = phi,
                               power = power)
    pmf <- 1-sum(temp)+temp[1]
  }
  if(method == "MC") {
    pmf <- monte_carlo(integrand = integrand, n_pts = n_pts, 
                       y = y, mu = mu, phi = phi, power = power)
  }
  return(pmf)
}
# Vectorize version
dptweedie <- Vectorize(FUN = dptweedie_aux, vectorize.args = c("y","mu"))


# Numerical integration using Monte Carlo method (PET) -----------------------
monte_carlo_PET <- function(y, mu,power, phi,n_pts,method,pp) 
{
  T <- length(y)
  f <- rep(NA,T)
  pts <- rexp(pp,rate=1)
  Z=rep(T,pp)
  for (i in 1:T){
    for (j in 1:pp){
      Z[j] <- dptweedie(y[i], mu = mu[i]*pts[j], phi = phi*((pts[j])^(1-power)), power = power,n_pts=n_pts,method=method)
    }
    x <- Z
    if (!is.data.frame(x)) x <- as.data.frame(x)
    # Les variables a traiter sont specifiees ?
    if (is.null(v))
      v <- names(x)
    # Pour chaque variable specifie, suppression des observations si NA
    Z1 <- x[complete.cases(x[v]),] 
    f[i] <- mean(Z1)
  }
  return(f)
}

disp_index_PET<- function(mu, phi, power) {
  vv <- (mu^2 + phi*mu^power+mu)/mu
  return(vv)
}

disp_index_PET_NB<- function(mu, phi, power) {
  vv <- (mu^2 + phi*mu^power+mu)/(mu+mu^2)
  return(vv)
}

